/**
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */
package org.apache.geronimo;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.URL;
import java.util.concurrent.ExecutorService;
import java.util.concurrent.Executors;
import java.util.concurrent.ThreadPoolExecutor;

import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;

import org.apache.ahc.AsyncHttpClient;
import org.apache.ahc.codec.HttpRequestMessage;
import org.apache.ahc.codec.HttpResponseMessage;
import org.apache.ahc.codec.SessionCache;
import org.apache.ahc.util.MonitoringEvent;
import org.apache.ahc.util.TimeMonitor;
import org.apache.catalina.CometEvent;
import org.apache.catalina.CometProcessor;

/**
 * Servlet implementation class for Servlet: AsyncServlet
 * 
 * @web.servlet name="AsyncServlet" display-name="AsyncServlet"
 * 
 * @web.servlet-mapping url-pattern="/async"
 * 
 */
public class AsyncServlet extends HttpServlet implements CometProcessor {
    @Override
    public void destroy() {
        super.destroy();
        ahc.destroyAll();
        threadPool.shutdownNow();
    }

    static final long serialVersionUID = 1L;

    private String remoteUrl;
    private static AsyncHttpClient ahc;
    private static ExecutorService threadPool;
    private static SessionCache sessionCache; 
    private static TimeMonitor monitor;

    public void init() throws ServletException {
        remoteUrl = getServletConfig().getInitParameter("remoteUrl");

        threadPool = Executors.newCachedThreadPool();
        sessionCache = new SessionCache(); 

        ahc = new AsyncHttpClient(threadPool, null);
        // enable connection reuse
        ahc.setSessionCache(sessionCache); 
        // disable TCP delay
        ahc.setTcpNoDelay(true);
        // enable the monitor
        monitor = new TimeMonitor();
        ahc.addMonitoringListener(monitor);
    }
    
    /**
     * Process the given Comet event.
     * 
     * @param event
     *            The Comet event that will be processed
     * @throws IOException
     * @throws ServletException
     */
    public void event(CometEvent event) throws IOException, ServletException {
        
        HttpServletResponse eventResponse = event.getHttpServletResponse();
        HttpServletRequest request = event.getHttpServletRequest();
        boolean debug = Boolean.parseBoolean(request.getParameter("debug"));
        boolean shortCircuit = Boolean.parseBoolean(request.getParameter("shortCircuit"));
        String cometId = null;
        
        if (event.getEventType() == CometEvent.EventType.BEGIN) {
            if (!shortCircuit) {
    
                cometId = request.getSession(true).getId();
                if(debug) {
                    System.out.println();
                    System.out.println("Event type: " + event.getEventType().toString() + " Session ID: " + cometId + "\n Request:" + request + " Response: " + eventResponse);
                }
                boolean dumpStats = request.getParameter("dump") != null;
                if (dumpStats) {
                    dumpStats(event.getHttpServletResponse());
                    return;
                }
                
                event.getHttpServletRequest().setAttribute(
                        "org.apache.tomcat.comet.timeout", new Integer(15 * 1000));
    
                // TODO this doesn't look thread safe...
                if (request.getParameter("remoteUrl") != null) {
                    remoteUrl = request.getParameter("remoteUrl");        
                }
    
                String remoteUri = request.getParameter("remoteUri");
                if (remoteUri == null) {
                    remoteUri = "/remoteApp/page";
                }
                
                URL url_connect = new URL(remoteUrl + remoteUri);       
                eventResponse.flushBuffer();
                Callback callback = new Callback(eventResponse, cometId, debug);
                //Callback callback = new Callback(event, cometId);
                HttpRequestMessage msgRequest = new HttpRequestMessage(url_connect, callback);
    
                // If the file being served has other URIs in it, like say images,
                // subsequent calls are made to the remote app to retrieve those.
                // We don't want to set new delay times and status codes then.
                if (remoteUri.endsWith("page")) {
                    msgRequest.setParameter("sessionId", cometId);
                    
                    int delayInt = 0;
                    String delay = request.getParameter("delay");
                    if (delay != null) {
                        delayInt = Integer.parseInt(delay);
                    } else {
                        delayInt = Utils.generateRandomDelay();
                        delay = String.valueOf(delayInt);
                    }
                    
                    String code = request.getParameter("code") == null ?
                            String.valueOf(Utils.generateRandomCode()) : 
                            request.getParameter("code");
                    
                    msgRequest.setParameter("delay", delay);
                    msgRequest.setParameter("code", code);
                    msgRequest.setTimeOut(delayInt + 15000);
                    if (debug) {
                        System.out.println("sessionId: " + cometId + "; delay: "
                            + delay + "; code: " + code);
                    }
                } else {
                    System.out.println("msg request is " + msgRequest.getPath());
                }
    
                msgRequest.setRequestMethod(HttpRequestMessage.REQUEST_GET);            
                
                try {
                    callback.setTime(System.nanoTime());
                    ahc.sendRequest(msgRequest);
                    //System.out.println("sessionId: " + cometId + " with request " + msgRequest + " sent");
                } catch (Exception e) {
                    e.printStackTrace();
                    // send an error response using the callback
                    callback.onException(e);
                    event.close();
                }
            } else {                
                synchronized(eventResponse) {
                    event.close();
                }
            }
        } else if (event.getEventType() == CometEvent.EventType.ERROR) {
            if (debug) {    
                System.out.println("Event type: " + event.getEventType().toString() + " Session ID: " + cometId + "\n Request:" + request + " Response: " + eventResponse);
            }
            synchronized(eventResponse) {
                event.close();
            }
        } else if (event.getEventType() == CometEvent.EventType.END) {
            if (debug) {
                System.out.println("Event type: " + event.getEventType().toString() + " Session ID: " + cometId + "\n Request:" + request + " Response: " + eventResponse);
            }
            synchronized(eventResponse) {
                event.close();
            }
        }
    }
    
    private void dumpStats(HttpServletResponse response) throws IOException, ServletException {
        response.setStatus(HttpServletResponse.SC_OK);
        response.setContentType("text/plain");
        String output = String.format("largest pool size: %d", 
                ((ThreadPoolExecutor)threadPool).getLargestPoolSize());
        PrintWriter out = response.getWriter();
        out.println(output);
        // print out the time monitor stats
        for (int i = 0; i <= MonitoringEvent.CONNECTION_CLOSED; i++) {
            out.println(i + ": " + monitor.getCount(i));
        }
        out.println("response time: " + monitor.getAverageResponseTime());
        out.println("connect time: " + monitor.getAverageConnectTime());
        out.close();
    } 
}
